# -*- coding: utf-8 -*-
"""
Functions related to timeseries management.

This file can also be imported as a module and contains the following
functions:

    * format_to_timetable - converts wind speed to power output of a wind turbine
    * synchronise - convert solar irradiance to the power output of a PV installation
    * resample_timeseries - resamples a timeseries to a given resolution
    * split_in_daily_profiles - splits a timeseries in a set of daily profiles
    * timestep_start - determines a timestep start for a given resolution
    * datetime_to_microepoch - converts a datetime to microepoch
    * microepoch_to_datetime_index - converts microepoch to datetime
    * microepoch_to_local_datetime - converts microepoch to a local datetime
    
"""

# Import all useful libraries
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
from dateutil.parser import parse
from dateutil.tz import tzlocal
from pytz import timezone

# Helpers
def __is_timezone(var):
    class_name = str(var.__class__)
    return class_name.startswith("<class 'pytz.tzfile.") or class_name.startswith(
        "<class 'dateutil.tz.tz."
    )


def __get_local_timezone():
    return tzlocal()


# Dealing with timeseries data
def format_to_timetable(data, time_column="t", column_dict=None):
    """Formats a pandas dataframe into a timetable

    Parameters
    ----------
    data : pd.DataFrame
        Dataframe to convert to timetable
    time_column : str
        Name of the column to use as time index
        (default: "t")
    column_dict : dict
        Dictionary to use for column renaming, if needed
        (default: None)

    Raises
    ------
    None

    Returns
    -------
    data : pd.DataFrame
        DataFrame indexed by time (timetable)
    """

    if len(data) == 0:
        return data

    if column_dict is not None:
        data = data.rename(columns=column_dict)

    if time_column != "t":
        data = data.rename(columns={time_column: "t"})

    if isinstance(data["t"][0], str):
        data["t"] = pd.to_datetime(data["t"])

    data = data.set_index("t").sort_index(ascending=True)
    return data


def synchronise(df1, df2, base="first", fill=None):
    """Formats a pandas dataframe into a timetable

    Parameters
    ----------
    df1 : pd.DataFrame
        First DataFrame to merge
    df2 : pd.DataFrame
        First DataFrame to merge
    base : str
        Time base for the merged DataFrame
        (options: "first","second","intersection","union")
        (default: "first")
    fill : str
        Filling method, if relevant
        (options: "backfill")
        (default: None)

    Raises
    ------
    None

    Returns
    -------
    data : pd.DataFrame
        DataFrame indexed by time (timetable)
    """

    # Some more hints here: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#resampling

    if base == "first":
        df_m = df1.merge(df2, how="left", right_index=True, left_index=True)

    elif base == "second":
        df_m = df1.merge(df2, how="right", right_index=True, left_index=True)

    elif base == "intersection":
        df_m = df1.merge(df2, how="intersection", right_index=True, left_index=True)

    elif base == "union":
        df_m = df1.merge(df2, how="outer", right_index=True, left_index=True)

    else:
        raise Exception("This case is not supported : " + base)

    if fill is not None:
        if fill == "backfill":
            df_m = df_m.fillna("backfill", axis="index")
        else:
            raise Exception("This fill is not supported yet: " + base)

    return df_m


def resample_timeseries(data, resolution="raw", function="mean", timezone=None):
    """Resamples a timeseries at a given frequency

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame to resample
    resolution : str
        Resolution to use for the daily profile
        (options: "raw","hour","day")
        (default: "raw")
    function : str
        Function to use for the resampling
        (options: "mean","min","max","sum")
        (default: "mean")

    Raises
    ------
    None

    Returns
    -------
    T_split : pd.DataFrame
        DataFrame of containing the split of the daily profiles, indexed by date
    """

    need_to_reset_timezone = False

    if resolution == "raw" or data is None or len(data) == 0:
        None

    else:
        if resolution == "hour":
            resampler = data.resample("H")
        elif resolution == "day":
            if timezone is None:
                resampler = data.resample("D")
            else:
                # Averaging per day is made in local timezone
                data.index = data.index.tz_convert(timezone)
                resampler = data.resample("D")
                need_to_reset_timezone = True

        else:
            raise Exception("Illegal value of data resolution: " + resolution)

        if function == "mean":
            data = resampler.mean()
        elif function == "sum":
            data = resampler.sum()
        elif function == "min":
            data = resampler.min()
        elif function == "max":
            data = resampler.max()
        else:
            raise Exception("Illegal value of function: " + function)

    if need_to_reset_timezone is True:
        data.index = data.index.tz_convert("UTC")

    return data


def split_in_daily_profiles(data, column, resolution="hour", time_zone="UTC"):
    """Formats a pandas dataframe into a timetable

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame to split in daily profiles
    column : str
        Name of the column to split in daily profiles
    resolution : str
        Resolution to use for the daily profile
        (options: "hour","15min","10min","5min")
        (default: "hour")
    time_zone : str or pytz timezone
        Time-zone for the daily profile
        (default: "UTC")

    Raises
    ------
    None

    Returns
    -------
    T_split : pd.DataFrame
        DataFrame of containing the split of the daily profiles, indexed by date
    """
    data_copy = data[[column]].copy()
    data_copy["date"] = data_copy.index.date

    if isinstance(time_zone, str):
        target_tz = timezone(time_zone)
    elif __is_timezone(time_zone):
        target_tz = time_zone
    else:
        raise Exception(
            "Invalid format for input 'time_zone' (must be either str or pytz timezone)"
        )

    data_copy.index = data_copy.index.tz_convert(target_tz)

    T_split = (
        pd.DataFrame({"date": data_copy.index.date})
        .drop_duplicates(subset=["date"])
        .set_index("date")
    )

    if resolution == "hour":
        data_copy["timeOfDay"] = data_copy.index.hour
        range2use = range(0, 24, 1)
        col_name = lambda x: "H" + str(x)

    elif resolution == "15min":
        data_copy["timeOfDay"] = data_copy.index.hour * 60 + data_copy.index.minute
        range2use = range(0, 24 * 60, 15)
        col_name = lambda x: "M15_" + str(int(x / 15))

    elif resolution == "10min":
        data_copy["timeOfDay"] = data_copy.index.hour * 60 + data_copy.index.minute
        range2use = range(0, 24 * 60, 10)
        col_name = lambda x: "M10_" + str(int(x / 10))

    elif resolution == "5min":
        data_copy["timeOfDay"] = data_copy.index.hour * 60 + data_copy.index.minute
        range2use = range(0, 24 * 60, 5)
        col_name = lambda x: "M5_" + str(int(x / 5))

    else:
        raise Exception("Illegal resolution: " + resolution)

    for i in range2use:
        T_i = (
            data_copy[[column, "date"]]
            .loc[data_copy["timeOfDay"] == i]
            .set_index("date")
        )
        T_i = T_i.rename(columns={column: col_name(i)})
        T_split = T_split.merge(T_i, how="left", right_index=True, left_index=True)

    return T_split


def split_data_loading_range(start, end, step=timedelta(weeks=24)):
    date_range = []
    max_date = lambda x: np.min([end, x + step])
    time_1 = start
    time_2 = start
    while time_2 < end:
        time_1 = time_2
        time_2 = max_date(time_1)
        date_range.append([time_1, time_2])

    return date_range


def timestep_start(step, t=None):

    if t is None:
        t = datetime.now(__get_local_timezone())

    if step == "second" or step == "1s":
        t_start = t.replace(microsecond=0)
    elif step == "15s":
        sec_start = int(t.second / 15) * 15
        t_start = t.replace(microsecond=0, second=sec_start)
    elif step == "30s":
        sec_start = int(t.second / 30) * 30
        t_start = t.replace(microsecond=0, second=sec_start)
    elif step == "minute" or step == "1min":
        t_start = t.replace(microsecond=0, second=0)
    elif step == "5min":
        min_start = int(t.minute / 5) * 5
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "15min":
        min_start = int(t.minute / 15) * 15
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "30min":
        min_start = int(t.minute / 30) * 30
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "hour":
        t_start = t.replace(microsecond=0, second=0, minute=0)
    elif step == "day":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0)
    elif step == "month":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0, day=1)
    elif step == "year":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0, day=1, month=1)
    else:
        raise Exception("Unknown setp: " + step)

    return t_start


# Functions dealing with microepoch data
def datetime_to_microepoch(datetimes2use):
    return int(datetimes2use.timestamp() * 1e6)


def microepoch_to_datetime_index(microepoch):
    t = pd.DatetimeIndex(np.round(microepoch).astype("datetime64[us]"))
    return t.tz_localize("UTC")


def microepoch_to_local_datetime(microepoch):
    t = pd.Datetime(np.round(microepoch).astype("datetime64[us]"))
    return t.tz_localize(__get_local_timezone())



# -*- coding: utf-8 -*-
"""
Functions related to figures and plots.

functions:

    * clean_legend - removes duplicate names in a legend

"""

import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.express.colors as pxc
from plotly.colors import n_colors

import numpy as np


def __make_color_scale(n, type="green-red"):

    if type == "green-red":
        colors = n_colors("rgb(5, 200, 200)", "rgb(200, 10, 10)", n, colortype="rgb")
    else:
        raise Exception("Unknown color scale: ")

    return colors


def __make_color_map(names, colormap, cycle_colormap=True):

    n_names = len(names)
    lower_cm = colormap.lower()

    if lower_cm == "plotly":
        colormap2use = pxc.qualitative.Plotly
    elif lower_cm == "d3":
        colormap2use = pxc.qualitative.D3
    elif lower_cm == "safe":
        colormap2use = pxc.qualitative.Safe
    elif lower_cm == "plasma_r":
        colormap2use = pxc.sequential.Plasma_r
    elif lower_cm == "green-red":
        colormap2use = __make_color_scale(n=n_names, type="green-red")
    else:
        raise Exception("Unknown color map: " + colormap)

    n_colors = len(colormap2use)

    if cycle_colormap is False and n_names > n_colors:
        raise Exception(
            "The chosen colormap ({}) is too short for your legend".format(colormap)
        )

    colors = dict()
    i = 0
    for name_i in names:
        colors[name_i] = colormap2use[np.mod(i, n_colors)]
        i = i + 1

    return colors


def __remove_duplicate_legends(fig):
    names_already_in = list()
    for sp_i in fig["data"]:
        if sp_i["name"] in names_already_in:
            sp_i["showlegend"] = False
        else:
            sp_i["showlegend"] = True
            names_already_in.append(sp_i["name"])


def __update_legend_colors(fig, colors):
    if isinstance(colors, str):
        # Identify all names
        all_names = list()
        for sp_i in fig["data"]:
            if sp_i["name"] not in all_names:
                all_names.append(sp_i["name"])

        # Create the color map
        colors = __make_color_map(all_names, colors)

    for sp_i in fig["data"]:
        sp_i_name = sp_i["name"]
        if sp_i_name in colors.keys():
            color2use = colors[sp_i_name]
            try:
                sp_i["line"]["color"] = color2use
            except:
                None

            try:
                sp_i["marker"]["color"] = color2use
            except:
                None


def __remove_all_grids(fig, background="#FFF"):

    if isinstance(fig["layout"], list):
        layouts = fig["layout"]
    else:
        layouts = [fig["layout"]]

    for lay_i in layouts:
        lay_i["plot_bgcolor"] = background

        for field_j in lay_i:
            if field_j.startswith("xaxis") or field_j.startswith("yaxis"):
                lay_i[field_j]["showgrid"] = False
                lay_i[field_j]["linecolor"] = "#BCCCDC"


def group_legend_by_name(fig, colors=None):
    """
    Groups alls legends of a figure according to their name (including subplots).

    Parameters
    ----------
    fig : (Figure)
        Figure identifier
    colors: dict, str or None
        Mapping of signal names to colors, or colormap to use
        Default: None
    """

    group_names = dict()

    for sp_i in fig["data"]:
        sp_i_name = sp_i["name"]
        if sp_i_name in group_names.keys():
            sp_i["legendgroup"] = group_names[sp_i_name]

        else:
            if sp_i.legendgroup is None:
                group_names[sp_i_name] = "plot_key_" + sp_i_name
            else:
                group_names[sp_i_name] = sp_i["legendgroup"]

            sp_i["legendgroup"] = group_names[sp_i_name]

    if colors is not None:
        __update_legend_colors(fig, colors)


def minimalistic_show(fig, renderer=None):
    """
    Makes a minimalistic display of a figure, by removing the following:
        - Background (white)
        - Grid lines
        - Tips
        - Display mode bar
        - Harmonizing legends (according to name)

    Parameters
    ----------
    fig : (Figure)
        Figure identifier
    renderer: str or None
        Renderer to use for the visualisation (None=default renderer)
        Default: None
    """
    __remove_all_grids(fig, background="#FFF")

    clean_legend(fig, colors="safe")

    if renderer is None:
        fig.show(config={"displayModeBar": False, "showTips": False})
    else:
        fig.show(renderer=renderer, config={"displayModeBar": False, "showTips": False})


def clean_legend(fig, colors=None):
    """
    Removes duplicate "name" in the legend of a figure (including subplots).

    Parameters
    ----------
    fig : (Figure)
        Figure identifier
    colors: dict or None
        Mapping of signal names to colors
        Default: None
    """

    # Join corresponding legends in groups
    group_legend_by_name(fig, colors)

    # Remove duplicates
    __remove_duplicate_legends(fig)



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  3 21:22:56 2020

@author: Dtime_PVF
"""

from datetime import datetime, timedelta
from seedftw.base.timeseries import split_data_loading_range

# Loading data from the API for 1 parameter
def split_api_calls(func, start, end, max_step, margin_last=timedelta(seconds=1)):
    """
    Function to split loading of data in a smart manner
    """

    # Create the steps for loading
    date_steps_range = split_data_loading_range(start=start, end=end, step=max_step)

    data = None
    for start_j, end_j in date_steps_range:
        data_j = func(start_j, end_j - margin_last)

        if data is None:
            data = data_j
        else:
            data = data.append(data_j)

    return data



# -*- coding: utf-8 -*-
"""
Functions related to data loading in Great Britain.

Sources:
- https://carbonintensity.org.uk/

"""

import pandas as pd
import requests
from datetime import datetime, timedelta
from pytz import timezone, UTC
from dateutil.parser import parse

from seedftw.base.timeseries import (
    format_to_timetable,
    resample_timeseries,
    timestep_start,
)
from seedftw.base.tricks import split_api_calls

# For caching
from functools import lru_cache

__uk_timezone = "Europe/London"
__api_url = "https://api.carbonintensity.org.uk/"

# Loading data from CarbonIntensity.org.uk
@lru_cache(maxsize=10)
def __api_call(endpoint):
    r = requests.get(__api_url + endpoint)

    data = r.json()

    if "error" in data.keys():
        status = data["error"]["code"]
    else:
        status = 200

    if status == 200:
        None
    elif status.startswith("400"):
        raise Exception("Bad request (on endpoint: {})".format(endpoint))
    elif status.startswith("500"):
        raise Exception("Internal server error (on endpoint: {})".format(endpoint))
    else:
        raise Exception(
            "Unknown return status={} (on endpoint: {})".format(status, endpoint)
        )

    return data


def __convert_string_to_datetime(string):
    return parse(string)


def get_emission_factors(out="dict"):
    endpoint = "intensity/factors"
    r = __api_call(endpoint)
    data = r["data"][0]

    low_format = out.lower()

    if low_format == "dict":
        return data
    elif low_format == "dataframe":
        return pd.DataFrame(
            {"technology": data.keys(), "emission_intensity": data.values()}
        )
    else:
        raise Exception("Unknown format: " + out)


def __load_historical_intensity(start, end):
    endpoint = "intensity/{}/{}".format(start.isoformat(), end.isoformat())
    r = __api_call(endpoint)
    raw_data = pd.DataFrame(r["data"])
    raw_data["start"] = pd.to_datetime(raw_data["from"])
    raw_data[["forecast", "actual", "index"]] = raw_data["intensity"].apply(pd.Series)

    raw_data2 = pd.DataFrame(raw_data["intensity"])

    data = format_to_timetable(
        raw_data[["start", "forecast", "actual"]], time_column="start", column_dict=None
    )
    return data


def get_historical_intensity(
    start=timestep_start("30min") - timedelta(days=2),
    end=timestep_start("30min"),
    resolution="raw",
):
    max_step = timedelta(days=13)
    data = split_api_calls(
        __load_historical_intensity,
        start,
        end,
        max_step,
        margin_last=timedelta(seconds=1),
    )

    data = resample_timeseries(
        data, resolution=resolution, function="mean", timezone=__uk_timezone
    )
    return data
